//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* curve.cpp: operations on curve files                                   *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.mesen-und-deuten.de/frida                                   *//
//**************************************************************************//

#include <iostream>
#include <boost/shared_ptr.hpp>
#include <boost/regex.hpp>
#include <boost/format.hpp>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include <lmmin.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <ask_simple_value.h>
#include <readln.h>

using namespace std;
using boost::format;

#include "mystd.h"
#include "olf.h"
#include "mem.h"
#include "curve.h"
#include "var.h"
#include "expr.h"
#include "opr.h"
#include "xax_lex.h"


//***************************************************************************//
//* Local declarations                                                      *//
//***************************************************************************//

//! local functions (wouldn't suffice "static" ??)

namespace NCurveFile
{
    double NumericIntegral(
        POlc fc, int k, int j, double low, double hig );
    // parameters for numeric integration:
    double numint_bound_low=-INFINITY, numint_bound_hig=INFINITY;
    // tuning parameters:
    int numint_mode=4;
    double numint_epsabs=1e-10, numint_epsrel=1e-10;
};

bool allow_slow_conv = true;


//! NFitTune holds auxiliary parameters for the fit routine

namespace NFitTune {
    double Epsilon = 1.e-14;
    double Factor=100.;
    int nCall=200;
    int nPrint=0;
    int mFitMetric=1;
    double maxtime = 3;
};


//***************************************************************************//
//* COlc implementation                                                     *//
//***************************************************************************//


//! Set auxiliary parameters.
//! Also used when loading function from disk.

void COlc::setAux()
{
    weighing = COlc::_ERR;
    cv_epsabs = NCurveFile::numint_epsabs;
    cv_epsrel = NCurveFile::numint_epsrel;
}


//! Parse function.
//! Also used when loading function from disk.

void COlc::parseFunction( const string& _expr )
{
    evaMode = COlc::_EXPR;
    expr = _expr;
    user_xaxparse( expr.c_str(), &T );
    if( !T->has_dummy() )
        cerr << "WARNING: curve has no formal argument t\n";
}


//! Prompt for curve description, and initialize things.

void COlc::askCurve( const string& quest )
{
    uint np;
ask_again:
    string resp = sask( quest+" (or c,ci; h for help) ?" );
    if        ( resp=="h" ) {
        cout <<
            "curve definition:\n"
            "- either an arithmetic expression,\n"
        "    containing the running variable t,\n"
        "    for instance: p0+p1*t\n"
        "- or one of the following keywords:\n"
        "    c: external curve evaluation command (hc for help)\n"
        "    ci: external command followed by interpolation (hci for help)\n";
        goto ask_again;
    } else if ( resp=="hc" ) {
        cout <<
            "the external curve evaluation command must:\n"
            "- read from stdin a whitespace separated list of:\n"
            "  - 1*int:    the number of parameters M\n"
            "  - M*double: the parameters\n"
            "  - 1*int:    the number of data points N\n"
            "  - N*double: the data points\n"
            "- write to stdout a whitespace separated list of:\n"
            "  - N*double: N lines with curve values y\n"
//            "  - N*double: the curve values\n"
            "- write to error messages to stderr\n";
        goto ask_again;
    } else if ( resp=="hci" ) {
        cout <<
            "the external curve evaluation command must:\n"
            "- read from stdin a whitespace separated list of:\n"
            "  - 1*int:      the number of parameters M\n"
            "  - M*double:   the parameters\n"
            "- write to stdout a whitespace separated list of:\n"
            "  - N*2*double: N lines with pairs x y\n"
//            "  - N*2*double: a sequence of data pairs: x0 y0 x1 y1 x2 y2 ...\n"
            "- write to error messages to stderr\n";
        goto ask_again;
    } else if ( resp=="c" ) {
        evaMode = COlc::_SCR;
        expr = sask( "Curve evaluation command ?" );
        np = iask( "Number of parameters ?" );
    } else if ( resp=="ci" ) {
        evaMode = COlc::_SCR_INTP;
        expr = sask( "Curve evaluation command ?" );
        np = iask( "Number of parameters ?" );
    } else {
        parseFunction( resp );
        np = T->npar();
    }
    if( np>100 )
        throw "PLAUSIBILITY CHECK FAILED: unexpected number of parameters np = "
            + strg(np);
    for( uint ip=0; ip<np; ip++ )
        PCo.push_back( CCoord("p" + strg(ip), "") );
    setAux();
}


//! Evaluate a curve, vectorial version. Used in fitting and in plotting.

void COlc::curve_val_vec( vector<double>* ret, const vector<double>& vt,
                          uint k, uint j ) const
  // input: the function arguments vt[], and the context k, j
  // output: the curve evaluation result *ret[]
{
    ret->resize( vt.size() );
    CResult val;
    if ( evaMode==_EXPR ) {
        CContext ctx( k, j );
        if        ( !T->has_dummy() ) {  // curve does not depend on t
            T->tree_val( val, ctx );
            if ( val.vectorial )
                throw "BUG: tree_curve_val did not get scalar result";
            (*ret)[0] = val.r;
        } else {
            ctx.request_VT( &vt );
            T->tree_val( val, ctx );
            if ( !val.vectorial )
                throw "BUG: tree_curve_val did not get vectorial result";
            *ret = val.v;
        }
    } else if ( evaMode==_SCR || evaMode==_SCR_INTP ) {

        string par_fifonam = string("/tmp/par-") + getenv( "LOGNAME" );
        string ret_fifonam = string("/tmp/ret-") + getenv( "LOGNAME" );

        // TODO: choose name as: mystd::next_tmp_file( "/tmp/frida.%d" );

        mystd::system( "rm -f "+par_fifonam );
        mystd::system( "rm -f "+ret_fifonam );
        if ( mkfifo( par_fifonam.c_str(), 0666 ) )
            throw "SYSTEM ERROR cannot make " + par_fifonam;
        if ( mkfifo( ret_fifonam.c_str(), 0666 ) )
            throw "SYSTEM ERROR cannot make " + par_fifonam;
  
        string cmd = expr + " < " + par_fifonam + " > " + ret_fifonam + " &";
        system( cmd.c_str() );

        int par_fifo;
        // we use open instead of fopen or ofstream,
        // because we need non-blocking mode.
        if ( !( par_fifo = open( par_fifonam.c_str(), O_WRONLY ) ) )
            throw "SYSTEM ERROR cannot open " + par_fifonam;
        fcntl( par_fifo, F_SETFL, O_NONBLOCK);

        string out = str( format("%i ") % nPar() );
        write( par_fifo, out.c_str(), out.size() );
        for ( uint ip=0; ip<nPar(); ++ip ) {
            string out = str( format("%g ") % VC(j)->P[ip] );
            write( par_fifo, out.c_str(), out.size() );
        }
        if ( evaMode==_SCR ) {
            string out = str( format("%i ") % nPar() );
            write( par_fifo, out.c_str(), out.size() );
            for ( uint i=0; i<vt.size(); ++i ) {
                string out = str( format("%g ") % vt[i] );
                write( par_fifo, out.c_str(), out.size() );
            }
        }
        close( par_fifo );

        CSpec S;
        double x, y;
        FILE *F;
        if( !(F = fopen( ret_fifonam.c_str(), "r")) )
            throw "BUG: cannot read results of system call";

        if ( evaMode==_SCR ) {
            for ( uint i=0; i<vt.size(); ++i ) {
                if ( fscanf( F, "%lg\n", &y )!=1 )
                    throw "script returns no value for t["+
                        strg(i)+"]="+strg(vt[i]);
                S.push_xy( vt[i], y );
            }
        } else {
            while ( fscanf( F, "%lg %lg\n", &x, &y )==2 )
                S.push_xy( x, y );
            // interpolation:
            if ( S.size()<2 )
                throw "script returns only " + strg(S.size()) + " lines";
            cout << "DEB call intpol " << vt.front() << " .. " << vt.back() 
                 << "\n";
            S.intpol( vt, *ret );
        }
        fclose( F );
        
    } else
        throw "BUG: unexpected evaluation mode";
}


//! Dito, scalar version, to obtain isolated points.

double COlc::curve_val_sca( double farg, uint k, uint j ) const
{
    vector<double> vt(1, farg), res(1);
    curve_val_vec( &res, vt, k, j );
    return res[0];
}


//***************************************************************************//
//* Curve description and parametrization: set and show                     *//
//***************************************************************************//


//! Create curve files.

void NCurveFile::CreateFitcurve()
{
    NOlm::IterateD fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fc( new COlc );
    fc->askCurve( "Curve definition" );

    POld fd;
    while( (fd=fiter()) ) {
        POlc fout( fc->new_olc() );
        vector<double> P( fout->nPar(), 1.0 );
        fout->name = string("fit_") + fd->name;
        fout->kd = fiter.k();
        fout->xco = fd->xco;
        fout->yco = fd->yco;
        fout->ZCo = fd->ZCo;
        fout->RPar = fd->RPar;
        fout->lDoc.push_back( "cc " + fc->expr + " # z from " + fd->name );
        for (uint j=0; j<fd->nJ(); j++) {
            PCurve sout( new CCurve );
            sout->z = fd->VS(j)->z;
            sout->P = P;
            sout->fixed.resize( fout->nPar(), false );
            fout->V.push_back( sout );
        }
        NOlm::OloAdd( fout );
    }
}


//! Create just one "artificial" curve file with no data reference

void NCurveFile::CreateFreecurve()
{
    static int default_no=0;

    POlc fout( new COlc );
    fout->askCurve( "Curve definition" );

    fout->name = string("curve") + strg(default_no++);
    fout->xco = CCoord( "x", "" );
    fout->yco = CCoord( "y", "" );
    fout->kd = -1;
    fout->lDoc.push_back( "cca " + fout->expr );
    PCurve sout( new CCurve );
    sout->P.resize    ( fout->nPar(), 1.0 );
    sout->fixed.resize( fout->nPar(), false );
    fout->V.push_back( sout );
    NOlm::OloAdd( fout );
}


//! modify function for existing file

void NCurveFile::ChangeExpr()
{
    POlc ftmp( new COlc );
    ftmp->askCurve( "Replace curve definition by" );
        
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fc;
    while( (fc=fiter()) ) {
        uint nParOld = fc->nPar();
        fc->expr = ftmp->expr;
        fc->T = ftmp->T;
        fc->lDoc.push_back( "ccm " + ftmp->expr );
        uint nParNew = fc->nPar();
        fc->PCo.resize(nParNew);
        for( uint ip=nParOld; ip<nParNew; ip++ ){
            fc->PCo[ip] = CCoord("p" + strg(ip), "");
        }
        for (uint j=0; j<fc->nJ(); j++) {
            fc->VC(j)->P.resize( nParNew );
            fc->VC(j)->fixed.resize( nParNew, false );
            for( uint ip=nParOld; ip<nParNew; ip++ )
                fc->VC(j)->P[ip] = 1.0;
        }
    }
}


//! Set fit range restrictions.

void NCurveFile::ChangeRange()
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    string expr = sask(
        "Fit range restriction (expr to be evaluated as boolean) ?" );
        
    POlc fc;
    while( (fc=fiter()) ) {
        fc->range_expr = expr;
        fc->lDoc.push_back( "cr " + expr );
        if( expr=="" ) {
            fc->range_T = PTree();
            continue;
        }
        PTree T;
        user_xaxparse( expr.c_str(), &T );
        fc->range_T = T;
    }
}


//! Short for op0, op1, ...

void NCurveFile::SetAllPars(void)
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fin;
    while( (fin=fiter()) ) {
        for( int ip=0; ip<fin->nPar(); ++ip )
            NOperate::Pointwise( "p" + strg(ip) );
    }
}


//! Display parameter list for set of online curves.

void NCurveFile::ShowPar(void)
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fin;
    while( (fin=fiter()) )
        cout << fin->pInfoCat();
}


//! Explain weight.

string COlc::weight_str() const
{
    switch ( weighing ) {
    case _LIN:
        return "constant";
    case _LOG:
        return "logarithmic";
    case _ERR:
        return "with reciprocal variance";
    }
    return "BUG: unexpected weight";
}

//! Return parameter list for one online curve.

vector<string> COlc::pInfo() const
{
#define PINFO_WRD_SIZ 15
    char wrd[PINFO_WRD_SIZ];
    string lin;
    vector<string> ret;
    // file-wide info:
    ret.push_back( expr );
    if( kd!=-1 )
        ret.push_back( "refers to data: " + strg(kd) );
    if( kconv!=-1 )
        ret.push_back( "convolute with: " + strg(kconv) );
    ret.push_back( "weighing:       " + weight_str() );
    // z's and parameters:
    lin = "j   ";
    for ( uint iz=0; iz<nZ(); iz++ ) {
        snprintf( wrd, PINFO_WRD_SIZ, "z%i           ", iz );
        lin += wrd;
    }
    for ( uint ip=0; ip<nPar(); ip++ ) {
        snprintf( wrd, PINFO_WRD_SIZ, "p%d            ", ip );
        lin += wrd;
    }
    ret.push_back( lin );
    // z's and parameters per spectrum:
    for ( uint j=0; j<nJ(); j++ ) {
        lin = "";
        snprintf( wrd, PINFO_WRD_SIZ, "%-2i%c ", j,
                  ProtectedSpecs.contains(j) ? '/' : ' ' );
        lin += wrd;
        for ( uint iz=0; iz<nZ(); iz++ ) {
            snprintf( wrd, PINFO_WRD_SIZ, "%-12.7g ", V[j]->z[iz] );
            lin += wrd;
        }
        for ( uint ip=0; ip<nPar(); ip++ ) {
            snprintf( wrd, PINFO_WRD_SIZ, "%c%-12.7g ",
                      VC(j)->fixed[ip] ? 'x' : ' ', VC(j)->P[ip] );
            lin += wrd;
        }
        ret.push_back( lin );
    }
    return ret;
}

string COlc::pInfoCat() const
{
    vector<string> vs = pInfo();
    string ret = "";
    for( int i=0; i<vs.size(); ++i )
        ret += vs[i] + "\n";
    return ret;
}


//! Fix or unfix curve parameters.

void NCurveFile::SetFixed( bool fix )
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    static string sel;
    sel = wask( string(fix?"Fix":"Unfix") + " which parameters", sel );
    if( sel=="" )
        return;

    POlc f;
    while( (f=fiter()) ) {
        CList lis( sel, 0, f->nPar()-1 );
        for( uint iv=0; iv<lis.size(); ++iv ) {
            for( uint j=0; j<f->nJ(); ++j ){
                if ( f->ProtectedSpecs.contains(j) )
                    continue;
                f->VC(j)->fixed[lis.V[iv]] = fix;
            }
        }
    }
}


//! Set data file or convolution file.

void NCurveFile::SetFileReference( const string& which )
// which = "d"  set data file
//         "v"  set convolution file
//         "v-" switch convolution off
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    string expr;
    PTree T;
    if( which=="v-" )
        ; // no preparation
    else {
        if     ( which=="d" )
            expr = sask( "Fit data file" );
        else if( which=="v" )
            expr = sask( "Convolute with file" );
        else
            throw "BUG: unexpected which="+which;
        if (expr!="")
            user_xaxparse( expr.c_str(), &T );
    }

    POlc fc;
    while ( (fc=fiter()) ) {
        uint k = fiter.k();
        if     ( which=="v-" )
            fc->kconv = -1;
        else if( expr=="" ) {
            cout << format( "curve %-2d -> data %-2d" ) % k % fc->kd;
            if( fc->kconv!=-1 )
                cout << format( "; convolute with %-2d" ) % k % fc->kconv;
            cout << "\n";
        } else {
            uint lval;
            CContext ctx( k );
            T->tree_uival( &lval, ctx );
            if( lval>=NOlm::MOM.size() )
                throw "no data file " + strg(lval);
            if( ! P2D( NOlm::MOM[lval] ) )
                throw "file " + strg(lval) + " isn't a data file";
            if     ( which=="d" )
                fc->kd = lval;
            else if( which=="v" )
                fc->kconv = lval;
        }
    }
}


//! Set auxiliary parameters for numeric convolution.

void NCurveFile::SetConvTuningPars( string which )
{
    // per-file ???
    throw string( "convolution tuning parameters not operational" );
    
    static int mode = -1;
    static double epsabs = 1e-10, epsrel = 1e-10;

    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fc;
    if        ( which=="?" ) {
        while ( (fc=fiter()) ) {
            printf( "file %i:\n", fiter.k() );
            printf( "   cv  convolute with:   %i\n", fc->kconv );
            printf( "   cva eps_abs:          %g\n", fc->cv_epsabs );
            printf( "   cvr eps_rel:          %g\n", fc->cv_epsrel );
        }
    } else if ( which=="m" ) {
        mode = iask( "Convolution mode (-1: Riemann, 0..7: smooth..singular" );
        if( mode<-1 || mode>7 ){
            mode = -1;
            return;
        }
    } else if ( which=="a" ) {
        epsabs = dask( "Eps abs for convolution", epsabs );
        while ( (fc=fiter()) )
            fc->cv_epsabs = epsabs;
    } else if ( which=="r" ) {
        epsrel = dask( "Eps rel for convolution", epsrel );
        while ( (fc=fiter()) )
            fc->cv_epsrel = epsrel;
    }
}

//! Set auxiliary parameters (weighing, ..).

void NCurveFile::SetProperties( string which )
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fc;
    while ( (fc=fiter()) ) {
        if      (which=="wc")
            fc->weighing = COlc::_LIN;
        else if (which=="wl")
            fc->weighing = COlc::_LOG;
        else if (which=="wv")
            fc->weighing = COlc::_ERR;
        else
            throw string( "unknown property" );
    }
}


//***************************************************************************//
//* Least-squares fitting                                                   *//
//***************************************************************************//


//! Data transfer between frida and lm_minimize callback functions.

typedef struct {
    POld fd;
    POlc fc;
    uint k;
    uint j;
    double timeout;
} FitDatTyp;


//! Callback routine for use in lm_minimize: function evaluation.

void globalEvaluate( const double* par, int m_dat, const void *data, 
                     double* fvec, int *info )
// arguments (required by lmmin):
//    par   r/w  function parameters, to be optimized
//    m_dat r    number of data points
//    data  r    pointer to arbitrary data, here defined as FitDatTyp
//    fvec  w    vector of residues, to be minimized
//    info  w    can be use to terminate the minimization
{
    try {
        FitDatTyp *mydata;
        mydata = (FitDatTyp*) data;
        if( mydata->timeout && time(NULL) > mydata->timeout )
            throw string( "reached fit time out" );
        POlc fc = mydata->fc;
        POld fd = mydata->fd;
        PCurve c = fc->VC(mydata->j);
        
        uint ip = 0;
        for( uint i=0; i<fc->nPar(); ++i )
            if ( !(fc->VC(mydata->j)->fixed[i]) )
                c->P[i] = par[ip++];
        
        const PSpec s = fd->VS(mydata->j);
        vector<double> fit;
        fc->curve_val_vec( &fit, s->x, mydata->k, mydata->j );

        if ( fc->weighing==COlc::_ERR && s->dy.size()!=m_dat )
            throw string( "missing error bars for weighing" );

        for ( uint i=0; i<m_dat; i++) {
                
            if( fit[i]==INFINITY ) // TEMPORARY UNTIL exceptions can be
                                   // thrown everywhere
                throw string( "function evaluation return INFINITY" );

            if        ( fc->weighing==COlc::_LIN ) {
                fvec[i] = s->y[i] - fit[i];
            } else if ( fc->weighing==COlc::_LOG ) {
                if ( s->y[i] < 1e-99 ) {
                    throw string( "some data points too small, zero, or"
                                  " negative - incompatible with log weight" );
                } else if ( fit[i] < 1e-99 ) {
                    fvec[i] = 99; // giant penalty
                } else {
                    fvec[i] = log( s->y[i] / fit[i] );
                }
            } else if ( fc->weighing==COlc::_ERR ) {
                if ( s->dy[i] < 1e-99 )
                    throw string( "cannot weigh with dy "
                                  "(some are too small, zero, or negative" );
                fvec[i] = ( s->y[i] - fit[i] ) / s->dy[i];                
            }
        }
    } catch ( string& s ) {
        cerr << "\n" << s << "\n";
        *info = -1;
    }
}


//! Callback routine for use in lmmin: monitor fit progress.

void globalPrintout( int n_par, const double* par, int m_dat, const void *data, 
                     const double* fvec, 
                     int printflags, int iflag, int iter, int nfev )
// arguments (required by lmmin; see also globalEvaluate):
//       iflag : 0 (init) 1 (outer loop) 2(inner loop) -1(terminated)
//       iter  : outer loop counter
//       nfev  : number of calls to *evaluate
{
    if ( NFitTune::nPrint ) {
        printf( "fitmon (%i %i %i) %4i", printflags, iflag, nfev, iter );
        for( uint ip=0; ip<n_par; ++ip )
            printf( " %19.11g", par[ip] );
        double msd=0;
        for( uint i=0; i<m_dat; ++i )
            msd += fvec[i]*fvec[i];
        printf( " -> %19.9g\n", msd );
    }
}


//! Set auxiliary parameters. Directly called from main program.

void NCurveFile::SetFitTuningPars( string which )
{
    if      ( which=="?" )
        printf( "Fit setup:\n   nCall=%d\n   nPrint=%d\n   Factor=%g\n"
                "   Eps=%g\n   Metric=%i\n",
                NFitTune::nCall,
                NFitTune::nPrint,
                NFitTune::Factor,
                NFitTune::Epsilon,
                NFitTune::mFitMetric );
    else if ( which=="p" )
        NFitTune::nPrint = iask( "Print frequency", NFitTune::nPrint );
    else if ( which=="c" )
        NFitTune::nCall = iask( "Max no of calls", NFitTune::nCall );
    else if ( which=="f" )
        NFitTune::Factor = dask( "NFitTune::Factor", NFitTune::Factor );
    else if ( which=="e" )
        NFitTune::Epsilon = dask( "NFitTune::Epsilon", NFitTune::Epsilon );
    else if ( which=="m" ){
        // metric (0) is the integer indicator returned by lmmin;
        // it cannot be chosen here
        printf( "Fit metric:\n"
                "(1) coefficient of determination\n"
                "(2) mean deviation\n" );
        int ret = iask( "Option", NFitTune::mFitMetric );
        if( ret<1 || ret>CCurve::mQuality )
            throw string( "invalid option" );
        NFitTune::mFitMetric = ret;
    } else if ( which=="t" )
        NFitTune::maxtime = dask( "Max duration of fits (in s)",
                                  NFitTune::maxtime );
}


//! Fit.

void NCurveFile::Fit( bool _allow_slow_conv )
{
    PSpec D;
    PCurve C;
    vector<double> Par;
    lm_control_struct control;
    lm_status_struct status;
    FitDatTyp data;
    double ysum1, ysum2, ydev2;
    allow_slow_conv = _allow_slow_conv;

    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fc;
    while ( (fc=fiter()) ) {
        data.k = fiter.k();
        if( fc->evaMode==COlc::_EXPR && !fc->T->has_dummy() )
            throw string( "curve has no formal argument t" );
        if( fc->kd>=NOlm::MOM.size() )
            throw string( "number of data file out of range" );
        POld fd;
        if ( !(fd= P2D( NOlm::MOM[fc->kd] )) )
            throw "file " + strg(fc->kd) + " is not a data file";
        if ( fc->nJ() != fd->nJ() )
            throw string( "data file and curve file have "
                          "different number of spectra" );
        if ( !fc->nJ() )
            throw string( "BUG: empty curve and data file" );
        if ( fc->weighing==COlc::_ERR && !fd->VS(0)->dy.size() ) {
            cout << "INFO: no error bars => set linear weights\n";
            fc->weighing=COlc::_LIN;
        }
        if ( fc->ProtectedSpecs.size() )
            cout << "INFO: some spectra are protected\n";
        // fc->lDoc.push_back(
        //    "fitted to file " + strg(fc->kd) + " " + fd->name );

        if( fc->range_T ) {
            fd = POld( fd->new_old() );
            for ( uint j=0; j<fc->nJ(); j++ ) {
                PSpec ein = fd->VS(j);
                PSpec eout = PSpec( new CSpec );
                eout->copy_z_base( ein );
                vector<double> range( ein->size() );
                fc->range_T->tree_vec_val( &range, 0, fc->kd, j );
                bool with_dy = ein->dy.size();
                for ( uint i=0; i<ein->size(); i++ ) {
                    if ( range[i] ) {
                        if( with_dy )
                            eout->push_xyd( ein->x[i], ein->y[i], ein->dy[i] );
                        else
                            eout->push_xy( ein->x[i], ein->y[i] );
                    }
                }
                fd->V.push_back( eout );
            }
        }
        
        data.fc = fc;
        data.fd = fd;

        uint np = fc->nPar();
                
        control = lm_control_double;
        control.epsilon = NFitTune::Epsilon;
        control.stepbound = NFitTune::Factor;
        control.maxcall = NFitTune::nCall;

        for ( uint j=0; j<fc->nJ(); j++ ) {
            if ( fc->ProtectedSpecs.contains( j ) )
                continue;
            data.j = j;
            if( fiter.size()>1 )
                printf( "%3d", data.k );
            printf( "%3d", j );
            D = fd->VS(j);
            C = fc->VC(j);
            if ( !D->size() ){
                printf( " empty range\n" );
                continue;
            }

            // determine number of free parameters
            uint npfree = 0;
            for ( uint ip=0; ip<np; ++ip )
                if( !fc->VC(j)->fixed[ip] )
                    ++npfree;
            if (npfree<=0)
                continue;
            Par.resize(npfree);
            uint ipf=0;
            for ( uint ip=0; ip<np; ip++ ) {
                if ( !fc->VC(j)->fixed[ip] )
                    Par[ipf++] = C->P[ip];
            }
                                        
            if ( NFitTune::nPrint ) printf("\n");

            data.timeout = time(NULL) + NFitTune::maxtime;

            lmmin( npfree, &(Par[0]), D->size(), &data,
                   globalEvaluate, &control, &status,
                   globalPrintout );

            if( status.info==11 )
                throw string( "exception in fit function" );

            for ( uint i=0; i<npfree; i++ )
                printf(" %12g", Par[i]);

            C->Quality[0] = (double) status.info;

            // calculate coefficient of determination R^2:
            uint nd = fd->nPts(j);
            ysum1 = 0, ysum2 = 0;
            for( uint i=0; i<nd; ++i ){
                ysum1 += D->y[i];
                ysum2 += SQR( D->y[i] );
            }
            ydev2 = ysum2 - SQR( ysum1 )/nd; // = nd * <(y-<y>)^2>
            C->Quality[1] = ydev2!=0 ? 1 - SQR(status.fnorm) / ydev2 : -1;
                        
            // mean deviation:
            C->Quality[2] = SQR(status.fnorm) / nd;

            printf( " (%3d->%9.7g)",
                    status.nfev, C->Quality[NFitTune::mFitMetric] );

            // print status:
            printf( "%s\n", lm_shortmsg[ status.info ] );
        }
    }
    allow_slow_conv = true;
}


//***************************************************************************//
//* Generate integral file (extract parameters, integrate, ...)             *//
//***************************************************************************//


//! Data transfer between frida and gsl_function.

typedef struct {
    POlc fc;
    uint k;
    uint j;
} EvalDatTyp;


//! Generic callback routine for tree evaluation by gsl_function.

double myeval( double x, void* data )
{
    EvalDatTyp *mydata = (EvalDatTyp*) data;
    return mydata->fc->curve_val_sca( x, mydata->k, mydata->j );
}


//! display or change tuning parameters for numeric integration

void NCurveFile::SetIntTuningPars( string which )
{
    if      ( which=="?" )
        printf( "Numeric integration setup:\n"
                "   cna epsabs=%g\n"
                "   cnr epsrel=%g\n"
                "   cnm external integration mode: %d\n",
                numint_epsabs, numint_epsrel, numint_mode );
    else if ( which=="a" )
        numint_epsabs = dask( "Espilon abs", numint_epsabs );
    else if ( which=="r" )
        numint_epsrel = dask( "Epsilon rel", numint_epsrel );
    else if ( which=="m" ){
        cout << "Numeric integration modes (ignored for infinite intervals):\n"
            "   0: non-adaptive Gauss-Kronod-87 for very smooth functions\n"
            "   1: adaptive Gauss-Kronod-61 for smooth functions\n"
            "   ...\n"
            "   6: adaptive Gauss-Kronod-15 for not so smooth functions\n"
            "   7: adaptive for functions with singularities\n";
        numint_mode = iask( "Option", numint_mode );
    } else {
        throw "unexpected mode";
    }
}


//! Outer wrapper: dialog, loop over files and parameters

void NCurveFile::IntegralProperty()
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    POlc fin;

    if( NRead::stack_empty() )
        cout << "Generate integral file:\n"
            "   0 | 1 | ...         extract curve parameters\n"
            "   q0 | q1 | q2        extract fit metrics\n"
            "   i                   numeric integral\n"
            "   <list>              comma-separated list thereof\n"
            "   *                   extract all curve parameters\n"
            ;
    static string todo = "*";
    todo = wask( "Option", todo );

    if      ( todo==" " )
        return;
    else if ( todo=="*" ) {
        uint maxP = 0;
        while( (fin=fiter()) ){
            if( fin->nPar()>maxP )
                maxP = fin->nPar();
        }
        fiter.reset();
        if( maxP==0 )
            throw "no parameters in file(s)";
        todo = "";
        uint ip=0;
        while( 1 ){
            todo += strg(ip);
            if( ++ip==maxP )
                break;
            todo += ",";
        }
    }

    while( todo!="" ) {
        string::size_type ji = todo.find_first_of( ',' );
        string mode;
        if      ( ji==0 )
            throw "unexpected comma";
        else if ( ji==string::npos ) {
            mode = todo;
            todo = "";
        } else {
            mode = todo.substr( 0, ji );
            todo = todo.erase( 0, ji+1 );
        }

        uint ip;
        string label = mode;
        if ( mode=="i" ){
            if( NRead::stack_empty() )
                cout << "integration bounds: -inf,inf are allowed\n";
            numint_bound_low = dask( "Integrate from", numint_bound_low );
            numint_bound_hig = dask( "Integrate to", numint_bound_hig );
        } else if ( isdigit(mode[0]) ) {
            if( sscanf( mode.c_str(), "%i", &ip )!=1 )
                throw "invalid parameter " + mode;
            label = "p" + mode;
            mode = "p";
        }

        while( (fin=fiter()) ){
            if( ip>=fin->nPar() )
                continue;

            POld fout( new COld );
            PSpec S( new CSpec );

            uint nz = fin->ZCo.size();
            bool savable = ( nz > 0 );

            if( savable ){
                fout = POld( new COld( fin.get() ) );
                if( fout->name.substr(0,4)=="fit_" )
                    fout->name.replace( 0, 4, "" );
                fout->name.insert( 0, label + "_" );
                fout->xco = fin->ZCo.back();
                if ( mode=="p" )
                    fout->yco = fin->PCo[ip];
                else
                    fout->yco = CCoord( mode, "" );
                if( fout->ZCo.size()<1 )
                    throw string( "input has no z coord" );
                fout->ZCo.pop_back();
                fout->lDoc.push_back( "ci " + ( mode=="p" ? strg(ip) : mode ) );
                S->z = fin->VC(0)->z;
                S->z.pop_back();
            }

            for (uint j=0; j<fin->nJ(); j++) {

                // determine integral value
                double val;
                if      ( mode=="p" )
                    val = fin->VC(j)->P[ip];
                else if ( mode=="q0" )
                    val = fin->VC(j)->Quality[0];
                else if ( mode=="q1" )
                    val = fin->VC(j)->Quality[1];
                else if ( mode=="q2" )
                    val = fin->VC(j)->Quality[2];
                else if ( mode=="i" )
                    val = NumericIntegral( fin, fiter.k(), j,
                                           numint_bound_low, numint_bound_hig );
                else
                    throw "invalid mode " + mode;

                // print out ?
                if (!savable) {
                    cout << fin->name << ": " << label << " = " << val << "\n";
                    continue;
                }
                
                // store in integral spectrum
                S->push_xy( fin->VC(j)->z[nz-1], val );
                // move z values
                if (nz>=2 &&
                    (j+1)<fin->nJ() && 
                    fin->VC(j+1)->z[nz-2]!=fin->VC(j)->z[nz-2]) {
                    fout->V.push_back(S);
                    S->clear();
                    S->z = fin->VC(j+1)->z;
                    S->z.pop_back();
                }
            }
            if (savable) {
                fout->V.push_back(S);
                NOlm::OloAdd( fout );
            }
        }
        fiter.reset();
    }
}


//! Numeric integration of one curve.

double NCurveFile::NumericIntegral(
    POlc fc, int k, int j, double low, double hig )
{
    EvalDatTyp data;
    data.fc = fc;
    data.k = k;
    data.j = j;

    return mystd::Integrate( &myeval, &data, low, hig,
                             numint_mode, numint_epsabs, numint_epsrel );
}


//***************************************************************************//
//* Calculate cumulative distribution (needed for SPHERES PST simulation)   *//
//***************************************************************************//


//! Data transfer between frida and gsl_function.

typedef struct {
    POlc fc;
    uint k;
    uint j;
    double low;
    double hig;
} CumDisDatTyp;


namespace NCurveFile
{
    void DivideInterval( 
        vector<double> &y, const int ilow, const int ihig,
        CumDisDatTyp &data );
}

//! Calculate cumulative probability distribution on equidistant grid

/* Input:  `dis(x)`, a probability distribution.
   Output: `cum(r)`, the inverse of the cumulative probability distribution
     based on `dis`. It is implicitely understood that the abscissa of `cum`
     is an equidistant grid 2^m+1 points from 0 to 1.

   The resulting `cum` is used as follows. To draw random values x
   that are distributed according to `dis`:
   - Draw a random number r from a uniform distribution in the interval [0,1].
   - Calculate ilow=floor(r*2^m) and ihig = ilow+1.
   - Obtain x = cum(ilow) + (cum(ihig)-cum(ilow))*(r-ilow/2^m).
*/

void NCurveFile::CumulativeDistribution(void)
{
    NOlm::IterateC fiter;
    if( !(fiter.size()) )
        throw "no curve to operate on";

    static int nb = 3;
    nb = iask( "How many twofold divisions", nb );
    if( nb<1 ) return;

    static double low=-3, hig=+3;
    low = dask( "Lower bound for Phi^-1(r)", low );
    hig = dask( "Upper bound for Phi^-1(r)", hig );
    if( low>=hig ) return;

    POlc fin;
    while( (fin=fiter()) ){
        POld fout( new COld );
        fout->name = "Phi_" + fin->name;
        fout->xco = CCoord( "r", "" );
        fout->yco = CCoord( "Phi^-1(r)", "" );
        fout->ZCo = fin->ZCo;
        fout->lDoc = fin->lDoc;
        fout->lDoc.push_back( "cumulative distribution" );
        
        for( uint j=0; j<fin->nJ(); j++ ){
            PCurve C = fin->VC(j);
            PSpec S( new CSpec );
            S->z = C->z;
            S->x.resize( (1<<nb)+1 );
            for( int i=0; i<(1<<nb)+1; ++i )
                S->x[i] = ((double)i)/(1<<nb);
            S->y.resize( (1<<nb)+1 );
            S->y[0] = low;
            S->y[1<<nb] = hig;

            // recursive solution (acts on S.y):
            CumDisDatTyp data;
            data.fc = fin;
            data.k = fiter.k();
            data.j = j;
            DivideInterval( S->y, 0, 1<<nb, data );

            fout->V.push_back(S);
        }

        NOlm::OloAdd( fout );
    }
}


//! Calculate Phi^-1(r) by recursively dividing Phi-intervals.

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>

double IntervalDifference( double mid, void *data )
{
    const CumDisDatTyp *mydata = (const CumDisDatTyp*) data;
    int j = mydata->j, k = mydata->k;
    POlc fc = mydata->fc;
    return NCurveFile::NumericIntegral( fc, k, j,
                                        mydata->low, mid )
        -  NCurveFile::NumericIntegral( fc, k, j,
                                        mid, mydata->hig );
}

void NCurveFile::DivideInterval(
    vector<double> &Y, const int ilow, const int ihig,
    CumDisDatTyp &data )
{
    if( ihig-ilow<=1 )
        return; // end of recursion

    // starting values for intersection:
    double low = Y[ilow], hig = Y[ihig], mid = (hig+low)/2;
    
    // integration limits for IntervalDifferences:
    data.low = low; data.hig = hig;

    // overhead for GSL root finder:
    const gsl_root_fsolver_type *root_solver = gsl_root_fsolver_brent;
    gsl_root_fsolver *s = gsl_root_fsolver_alloc( root_solver );
    gsl_function F;
    F.function = &IntervalDifference;
    F.params = &data;
    gsl_root_fsolver_set( s, &F, low, hig );
    int status, iter = 0, max_iter = 20;

    // iterative root finding:
    do {
        //printf( "%4d %4d %5d [%.7f, %.7f] %.7f\n",
        //        ilow, ihig, iter, low, hig, mid );
        iter++;
        status = gsl_root_fsolver_iterate (s);
        mid = gsl_root_fsolver_root (s);
        low = gsl_root_fsolver_x_lower (s);
        hig = gsl_root_fsolver_x_upper (s);
        status = gsl_root_test_interval( low, hig, 0, 1e-10 );
        // if (status == GSL_SUCCESS) printf ("Converged.\n");
    } while( status == GSL_CONTINUE && iter < max_iter );
    gsl_root_fsolver_free (s);
   
    // save result:
    int imid = (ilow+ihig)/2;
    Y[imid] = mid;

    // continue recursion:
    DivideInterval( Y, ilow, imid, data );
    DivideInterval( Y, imid, ihig, data );
}
