//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* expr.h: expressions                                                    *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.mesen-und-deuten.de/frida                                   *//
//**************************************************************************//

#include <boost/shared_ptr.hpp>

#define PTree boost::shared_ptr<class CTree>
#define PRef  boost::shared_ptr<class CRef>
#define PRgr  boost::shared_ptr<class CRegRef>


//! A Context specifies how a tree shall be evaluated.

class CContext {
 public:
    enum TDim { _1, _VI, _VT } dim;
    uint k, j, i;
    bool want_error;
    uint nv;
    const vector<double> *vt;

    CContext( uint _k=-1, uint _j=-1, uint _i=-1 )
        : dim(_1), k(_k), j(_j), i(_i), want_error(false), nv(0), vt(0) {};

    void request_VT( const vector<double> *_vt ) {
        dim = _VT; nv = _vt->size(); vt = _vt; }
    void request_VI( uint _nv ) {
        dim = _VI; nv = _nv; }

    string context_info() const;
};


//! Return value for tree evaluation.

class CResult {
 public:
    bool vectorial;
    double r;
    double dr;
    vector<double> v;
    vector<double> dv;

    CResult() : vectorial(false), dr(0) {};

    void set_r( double _r ) { vectorial = false; r = _r; };
    void preset_v( int n ) { vectorial = true; v.resize(n); };
    void set_v( const vector<double>& v_ ) { vectorial = true; v = v_; };
    double to_r( int i ) const { return vectorial ? v[i] : r; };
    double to_dr( int i ) const
        { return vectorial ? ( dv.size() ? dv[i] : 0 ) : dr; };
    bool has_err() const { return vectorial ? dv.size() : dr; };

    string result_info() const;
};


//! References to data points, indices, ...; used as endpoints of CTree's.

class CRef: public CVariable {
 public:
    PTree tk;
    PTree tj;
    PTree ti;

    CRef( const string s="" ) : CVariable(s) {};
    
    void ref_val( CResult& ret, const CContext& ctx ) const;
    void set_coord( CCoord& ret, uint k ) const;

    string ref_info() const;
};


//! Parsed generic expression.

class CTree {
 private:
    enum TreeTyp {
        _NOTREE=0,
        _OP,    // operator (unary, binary, .. according to narg)
        _VAL,   // numeric value
        _CEV,   // curve evaluation 'c'
        _DUMMY, // curve argument 't'
        _REF,   // reference
        _RGR,   // register reference
        _CONV,  // convolution
        _DIRAC  // Dirac delta function
    };
    TreeTyp typ;
    double val;
    const class CFunc* fun;
    static const uint maxarg=3;
    PTree arg[maxarg];
    uint narg;
    PRef ref;
    PRgr rgr;

    void tree_op   ( CResult& ret, const CContext& ctx ) const;
    void tree_conv ( CResult& ret, const CContext& ctx ) const;
    void tree_cev  ( CResult& ret, const CContext& ctx ) const;

 public:
    CTree();
    CTree( double v_in );
    CTree( const class CFunc *f_in, PTree a0 );
    CTree( const class CFunc *f_in, PTree a0, PTree a1 );
    CTree( const class CFunc *f_in, PTree a0, PTree a1, PTree a2 );
    CTree( PRef& _ref );
    CTree( PRef& _ref, PTree a0 );
    CTree( PRgr& _rgr );
    CTree( char which );
    CTree( char which, const PTree& a0 );
    CTree( char which, const PTree& a0, const PTree& a1 );

    static void setConv( uint& kconv, uint& jconv, uint k, uint j );

    uint npar() const;
    bool k_dependent() const;
    bool has_dummy() const;

    void tree_val  ( CResult& ret, const CContext& ctx ) const;

    void tree_uival( uint *ret, const CContext& ctx ) const;
    void tree_point_val(
        double *ret, double *dret,
        uint k=(uint)-1, uint j=(uint)-1, uint i=(uint)-1 ) const;
    void tree_vec_val(
        vector<double> *ret, vector<double> *dret,
        uint k=(uint)-1, uint j=(uint)-1 ) const;

    void set_coord( CCoord& ret, uint k=(uint)-1 ) const;
    string tree_info() const;

 private:
    void npar_exec( uint *np ) const;
    void kdep_exec( bool *kd ) const;
};
