//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* zentry.cpp: z-dependent entries (spec or curve)                        *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.messen-und-deuten.de/frida                                  *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <vector>
#include <boost/shared_ptr.hpp>

using namespace std;

#include "mystd.h"
#include "zentry.h"


//**************************************************************************//
//*  class Zentry                                                          *//
//**************************************************************************//

//! Partial copy, to initialize a new Zentry from an existing one.

void CZentry::copy_z_base( PZentry ein )
{
    z = ein->z;
    dz = ein->dz;
}


//! Compare z vectors of two zentries, for use within sort routine.

bool CompareZ( const PZentry& E1, const PZentry& E2 )
{
    if( E1->z.size()!=E2->z.size() )
        throw string( "BUG: inconsistent z.size in CompareZ" );
    uint nz = E1->z.size();
    for( uint iz=0; iz<nz; ++iz ){
        if    ( E1->z[iz] > E2->z[iz] )
            return false;
        if    ( E1->z[iz] < E2->z[iz] )
            return true;
    }
    return false;
}


//**************************************************************************//
//*  class CSpec                                                           *//
//**************************************************************************//

//! Append given data point.

void CSpec::push_xy( double _x, double _y )
{
    x.push_back(_x);
    y.push_back(_y);
}


//! Append given data point, with error.

void CSpec::push_xyd( double _x, double _y, double _dy )
{
    if( y.size()!=dy.size() )
        throw string( "cannot push_xyd since y and dy have different size" );
    x.push_back(_x);
    y.push_back(_y);
    dy.push_back(_dy);
}


//! Remove data points according to given indices.

void CSpec::remove( uint from, uint to )
{
    x.erase( x.begin()+from, x.begin()+to );
    y.erase( y.begin()+from, y.begin()+to );
    if( dy.size() )
        dy.erase( dy.begin()+from, dy.begin()+to );
}


//! Remove unused dy.

void CSpec::purge_dy()
{
    for( uint i=0; i<dy.size(); ++i )
        if( dy[i] )
            return;
    dy.clear();
}


//! Reset to initial state.

void CSpec::clear()
{
    x.clear();
    y.clear();
    dy.clear();
    z.clear();
    dz.clear();
}


//! Resize x,y,dy.

void CSpec::resize( uint n, bool with_dy )
{
    x.resize( n );
    y.resize( n );
    if( with_dy ) dy.resize( n );
}


//! Return number of data points.

uint CSpec::size() const
{
    if ( x.size()!=y.size() )
        throw "BUG detected by spec::size: #x [" + 
                  strg(x.size()) + "] <> #y [" + strg(y.size()) + "]\n";
    if ( dy.size()!=y.size() && dy.size()!=0 )
        throw "BUG detected by spec::size: #y [" + 
                  strg(y.size()) + "] <> #dy [" + strg(dy.size()) + "]\n";
    return y.size();
}


//! For each xx[i], set yy[i] by interpolating in this.x -> this.y.

void CSpec::intpol( const vector<double>& xx, vector<double>& yy ) const
{
    uint n = size();
    if( n<2 )
        throw "BUG in intpol: less than two points to interpolate between";
    uint nn = xx.size();
    if( nn<1 )
        throw "BUG in intpol: empty vector of points to interpolate for";
    yy.resize( nn, 0. );

    uint ii = 0;
    while( xx[ii]<x[0] ){
        if( ++ii>=nn )
            return;
        // yy[ii] = 0;
    }
    uint i = 0;
    while ( i<n-1 && xx[ii]<x[n-1] ) {
        while ( xx[ii]>x[i+1] && i<n-1 )
            ++i;
        if( i>=n-1 )
            break;
        if( ii>=nn )
            return;
        yy[ii] = y[i] + (xx[ii]-x[i])/(x[i+1]-x[i])*(y[i+1]-y[i]);
        ++ii;
    }
    //for( ; ii<nn; ++ii )
    //    yy[ii] = 0;
}


//**************************************************************************//
//*  class CCurve                                                          *//
//**************************************************************************//

//! Initialize curve. Used when creating or loading a curve file.

void CCurve::clear(void)
{
    P.clear();
    z.clear();
    Quality[0] = -1;
    Quality[1] = -1;
    Quality[2] = -1;
}
